!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   rc_convert_mod
!> @brief   convert remd trajectory files
!! @authors Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module rc_convert_mod

  use rc_option_str_mod
  use pbc_correct_mod
  use fitting_mod
  use fitting_str_mod
  use trajectory_str_mod
  use output_str_mod
  use input_str_mod
  use select_atoms_mod
  use select_atoms_str_mod
  use molecules_str_mod
  use fileio_trj_mod
  use fileio_mod
  use string_mod
  use messages_mod
 
  implicit none
  private
  
  ! parameters
  integer, private, parameter :: MaxTrjFiles = 500

  ! structures
  type s_cnv_info
    integer              :: nreplicas
    integer              :: nsteps
    integer              :: rep_ex_period
    integer, allocatable :: ids(:)
    type(s_selatoms)     :: trjout_atom_trj

    integer              :: trj_file_idx
    type(s_trj_file)     :: trj_files(MaxTrjFiles)
  end type s_cnv_info

  ! subroutines
  public  :: convert
  private :: convert_replica
  private :: convert_parameter
  private :: check_convert_info
  private :: open_input_trj
  private :: open_input_trj_
  private :: close_input_trj
  private :: get_replicate_name

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert
  !> @brief        convert remd trajectory files
  !! @authors      NT
  !! @param[in]    input      : input information
  !! @param[in]    output     : output information
  !! @param[in]    option     : option information
  !! @param[inout] molecule   : molecule information
  !! @param[inout] trajectory : trajectory information
  !! @param[inout] fitting    : fitting information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine convert(input,      &
                     output,     &
                     option,     &
                     molecule,   &
                     trajectory, &
                     fitting)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(in)    :: option
    type(s_molecule),        intent(inout) :: molecule
    type(s_trajectory),      intent(inout) :: trajectory
    type(s_fitting),         intent(inout) :: fitting

    ! local variables
    type(s_cnv_info)         :: cnv_info


    ! check check only
    !
    if (option%check_only) &
      return

    if (input%dcdfile == '') &
      call error_msg('Run_Rmdcnv> input dcd file is empty.')

    if (input%remfile == '') &
      call error_msg('Run_Rmdcnv> input rem file is empty.')

    if (output%trjfile == '') &
      call error_msg('Run_Rmdcnv> output trj file is empty.')


    ! check information for convert
    !

    call check_convert_info(input, option, cnv_info)


    ! convert by..
    !
    write(MsgOut,'(A)') 'Run_Rmdcnv> convertion trajectory file.'
    write(MsgOut,'(A)') ' '

    select case (option%convert_type)

    case (ConvertTypeReplica)

      ! replica
      call convert_replica  (input,      &
                             output,     &
                             option,     &
                             molecule,   &
                             trajectory, &
                             fitting,    &
                             cnv_info)

    case default

      ! parameter
      call convert_parameter(input,      &
                             output,     &
                             option,     &
                             molecule,   &
                             trajectory, &
                             fitting,    &
                             cnv_info)

    end select

    write(MsgOut,'(A)') ' '


    ! deallocate
    !
    deallocate(cnv_info%ids)
    call dealloc_selatoms(cnv_info%trjout_atom_trj)

    return

  end subroutine convert

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert_replica
  !> @brief        convert remd trajectory files as replica IDs
  !! @authors      NT
  !! @param[in]    input      : input information
  !! @param[in]    output     : output information
  !! @param[in]    option     : option information
  !! @param[inout] molecule   : molecule information
  !! @param[inout] trajectory : trajectory information
  !! @param[inout] fitting    : fitting information
  !! @param[inout] cnv_info   : information for convert
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine convert_replica(input,      &
                             output,     &
                             option,     &
                             molecule,   &
                             trajectory, &
                             fitting,    &
                             cnv_info)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(in)    :: option
    type(s_molecule),        intent(inout) :: molecule
    type(s_trajectory),      intent(inout) :: trajectory
    type(s_fitting),         intent(inout) :: fitting
    type(s_cnv_info),        intent(inout) :: cnv_info

    ! local variables
    type(s_trj_file)         :: rep_in, rep_out
    integer                  :: i, j, rep_id, nids, nframes


    nids    = size(cnv_info%ids)
    nframes = cnv_info%nsteps/option%dcd_md_period


    do i = 1, nids

      rep_id = cnv_info%ids(i)

      write(MsgOut,'(A,I5, A,I5)') &
           'Id: ', i, '  Replica id: ', rep_id

      ! open trajectory files
      !

      call open_trj(rep_in,  &
                    get_replicate_name(input%dcdfile, rep_id),&
                    TrjFormatDCD,   &
                    TrjTypeCoorBox, &
                    IOFileInput)

      call open_trj(rep_out, &
                    get_replicate_name(output%trjfile, i), &
                    option%trjout_format, &
                    option%trjout_type,   &
                    IOFileOutputNew)

      do j = 1, nframes

        ! read trajectory
        !

        call read_trj(rep_in, trajectory)

        ! selection
        !

        call reselect_atom  (molecule,               &
                             option%trjout_atom_exp, &
                             trajectory%coord,       &
                             option%trjout_atom,     &
                             cnv_info%trjout_atom_trj)

        ! pbc-correct
        !

        call run_pbc_correct(option%pbcc_mode,       &
                             molecule,               &
                             trajectory)

        ! fitting
        !
        if (fitting%mass_weight) then
          call run_fitting  (fitting,                &
                             molecule%atom_coord,    &
                             trajectory%coord,       &
                             trajectory%coord,       &
                             molecule%mass)

        else
          call run_fitting  (fitting,                &
                             molecule%atom_coord,    &
                             trajectory%coord,       &
                             trajectory%coord)
        end if

        ! write trajectory
        !

        call write_trj(rep_out, trajectory, &
                       cnv_info%trjout_atom_trj, molecule)

      end do

      ! close trajectory files
      !

      call close_trj(rep_out)
      call close_trj(rep_in)

    end do

    return

  end subroutine convert_replica

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert_parameter
  !> @brief        convert remd trajectory files as parameter IDs
  !! @authors      NT, TM
  !! @param[in]    input      : input information
  !! @param[in]    output     : output information
  !! @param[in]    option     : option information
  !! @param[inout] molecule   : molecule information
  !! @param[inout] trajectory : trajectory information
  !! @param[inout] fitting    : fitting information
  !! @param[inout] cnv_info   : information for convert
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine convert_parameter(input,      &
                               output,     &
                               option,     &
                               molecule,   &
                               trajectory, &
                               fitting,    &
                               cnv_info)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(in)    :: option
    type(s_molecule),        intent(inout) :: molecule
    type(s_trajectory),      intent(inout) :: trajectory
    type(s_fitting),         intent(inout) :: fitting
    type(s_cnv_info),        intent(inout) :: cnv_info

    ! local variables
    type(s_trj_file)         :: rep_out
    integer                  :: i, frm_no, exc_no, rep_id, prm_id, nids, nframes

    type(s_trj_file), pointer :: rep_in

    
    nids    = size(cnv_info%ids)
    nframes = cnv_info%nsteps/option%dcd_md_period


    do i = 1, nids

      prm_id = cnv_info%ids(i)

      write(MsgOut,'(A,I5, A,I5)') &
           'Id: ', i, '  Parameter id: ', prm_id

      ! open output trajectory file
      !

      call open_trj(rep_out, &
                    get_replicate_name(output%trjfile, i), &
                    option%trjout_format, &
                    option%trjout_type,   &
                    IOFileOutputNew)

      exc_no = 0

      do frm_no = 1, nframes

        ! open input trajectory file
        !
        call open_input_trj(prm_id, frm_no, exc_no, rep_id, &
                            input, option, trajectory, cnv_info, rep_in)

        ! read trajectory
        !

        call read_trj(rep_in, trajectory)

        ! selection
        !

        call reselect_atom  (molecule,               &
                             option%trjout_atom_exp, &
                             trajectory%coord,       &
                             option%trjout_atom,     &
                             cnv_info%trjout_atom_trj)

        ! pbc-correct
        !

        call run_pbc_correct(option%pbcc_mode,       &
                             molecule,               &
                             trajectory)

        ! fitting
        !
        if (fitting%mass_weight) then
          call run_fitting  (fitting,                &
                             molecule%atom_coord,    &
                             trajectory%coord,       &
                             trajectory%coord,       &
                             molecule%mass)

        else
          call run_fitting  (fitting,                &
                             molecule%atom_coord,    &
                             trajectory%coord,       &
                             trajectory%coord)
        end if

        ! write trajectory
        !

        call write_trj(rep_out, trajectory, &
                       cnv_info%trjout_atom_trj, molecule)

      end do

      ! close trajectory files
      !

      call close_input_trj(cnv_info)
      call close_trj(rep_out)

    end do

    return

  end subroutine convert_parameter

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    check_convert_info
  !> @brief        check the information for conversion
  !! @authors      NT, TM
  !! @param[in]    input    : input information
  !! @param[in]    option   : option information
  !! @param[inout] cnv_info : information for conversion
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_convert_info(input, option, cnv_info)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_option),          intent(in)    :: option
    type(s_cnv_info),        intent(inout) :: cnv_info

    ! local variables
    integer                  :: i, nline, istep_tmp
    integer                  :: rem_io, stat, tmp
    character(MaxFilename)   :: remfile
    logical                  :: exist


    cnv_info%trj_file_idx   = 0
    cnv_info%trj_files%name = ''

    ! check number of replicas
    !
    remfile = input%remfile

    inquire(file=get_replicate_name(remfile,1), exist=exist)

    if (.not. exist) &
      call error_msg('Run_Rmdcnv> File not found. '//trim(remfile))

    cnv_info%nreplicas = 0
    do while (.true.)

      inquire(file=get_replicate_name(remfile,cnv_info%nreplicas+1), &
                                              exist=exist)
      if (.not. exist) &
        exit

      cnv_info%nreplicas = cnv_info%nreplicas + 1

    end do

    ! check number of MD steps
    !
    call open_file(rem_io, get_replicate_name(remfile,1), &
                   IOFileInput)

    nline = 0
    do while (.true.)
      read(rem_io,*,iostat=stat) cnv_info%nsteps, tmp
      if (stat /= 0) &
        exit
      nline = nline + 1

      if (nline == 1) then
        istep_tmp = cnv_info%nsteps
      else if (nline == 2) then
        cnv_info%rep_ex_period = cnv_info%nsteps - istep_tmp
      end if
    end do

    cnv_info%nsteps = cnv_info%nsteps + cnv_info%rep_ex_period

    write(MsgOut,'(A)')    'Check_Convert_Info>'
    write(MsgOut,'(A,I9)') '             # of replicas : ', cnv_info%nreplicas
    write(MsgOut,'(A,I9)') '             # of MD steps : ', cnv_info%nsteps
    write(MsgOut,'(A,I9)') '         DCD output period : ', option%dcd_md_period
    write(MsgOut,'(A,I9)') '   Replica exchange period : ', &
                                                          cnv_info%rep_ex_period
    write(MsgOut,'(A)')    ' '
    
    call close_file(rem_io)


    ! compute convert ids
    !

    if (.not. allocated(option%convert_ids)) then

      allocate(cnv_info%ids(cnv_info%nreplicas))
      do i = 1, cnv_info%nreplicas
        cnv_info%ids(i) = i
      end do

    else

      allocate(cnv_info%ids(size(option%convert_ids)))
      cnv_info%ids(:) = option%convert_ids(:)

      do i = 1, size(cnv_info%ids)
        if (cnv_info%ids(i) > cnv_info%nreplicas) &
          call error_msg('Check_Convert_Info> Convert IDs is out-of-range.')
      end do

    end if

    return

  end subroutine check_convert_info

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_input_trj
  !> @brief        open input trajectory file
  !! @authors      NT, TM
  !! @param[in]    prm_id     : parameter id
  !! @param[in]    frm_no     : trajectory frame number
  !! @param[inout] exc_no     : replica exchange number
  !! @param[in]    input      : input information
  !! @param[in]    option     : option information
  !! @param[inout] trajectory : trajectory information
  !! @param[inout] cnv_info   : information for convert
  !! @param[inout] rep_in     : trajectory file handle
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_input_trj(prm_id, frm_no, exc_no, rep_id, &
                            input, option, trajectory, cnv_info, rep_in)

    ! formal arguments
    integer,                 intent(in)    :: prm_id
    integer,                 intent(in)    :: frm_no
    integer,                 intent(inout) :: exc_no
    integer,                 intent(inout) :: rep_id
    type(s_input),           intent(in)    :: input
    type(s_option),          intent(in)    :: option
    type(s_trajectory),      intent(inout) :: trajectory
    type(s_cnv_info),        intent(inout) :: cnv_info
    type(s_trj_file),        pointer       :: rep_in

    ! local variables
    integer                  :: idcd, irex_b, irex_e, prm_id0
    integer                  :: i, rem_io, stat, tmp
    character(MaxFilename)   :: remfile, trjfile


    idcd   =  frm_no    * option  %dcd_md_period
    irex_b = (exc_no-1) * cnv_info%rep_ex_period
    irex_e =  exc_no    * cnv_info%rep_ex_period


    if (irex_e < idcd) then

      ! re-compute exchange section
      do while(.true.)
        exc_no =  exc_no + 1
        irex_b = (exc_no-1) * cnv_info%rep_ex_period
        irex_e =  exc_no    * cnv_info%rep_ex_period
        if (irex_b < idcd .and. idcd <= irex_e) &
          exit
      end do

      remfile = input%remfile

      ! find indicated parameters from rem files
      do rep_id = 1, cnv_info%nreplicas

        call open_file(rem_io, &
             get_replicate_name(remfile,rep_id), IOFileInput)

        do i = 1, exc_no
          read(rem_io,*,iostat=stat) tmp, prm_id0
          if (stat /= 0) &
            call error_msg('Open_input_trj> REM file read error.')
        end do

        call close_file(rem_io)

        if (prm_id == prm_id0) &
          exit

      end do
      if (rep_id > cnv_info%nreplicas) &
        call error_msg( &
        'Open_Input_Trj> Parameter ID is not found in remfile.')

      ! open input trajectory file
      trjfile = get_replicate_name(input%dcdfile, rep_id)

      call open_input_trj_(trjfile, frm_no, trajectory, cnv_info, rep_in)

    end if

    write(MsgOut,'(A,I9,A,I9)') '    MD step:', idcd, '  Replica:', rep_id

    return

  end subroutine open_input_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    open_input_trj_
  !> @brief        open and re-use input trajectory file
  !! @authors      NT
  !! @param[in]    filename   : filename
  !! @param[in]    frm_no     : trajectory frame index
  !! @param[inout] trajectory : trajectory information
  !! @param[inout] cnv_info   : information of convert
  !! @param[inout] rep_in     : handle for trajectory input file
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_input_trj_(filename, frm_no, trajectory, cnv_info, rep_in)

    ! formal arguments
    character(*),             intent(in)    :: filename
    integer,                  intent(in)    :: frm_no
    type(s_trajectory),       intent(inout) :: trajectory
    type(s_cnv_info), target, intent(inout) :: cnv_info
    type(s_trj_file),         pointer       :: rep_in

    ! local variables
    integer                   :: i, ti


    ! re-use
    !
    do i = 1, MaxTrjFiles

      ti  = mod(cnv_info%trj_file_idx, MaxTrjFiles) + 1

      if (cnv_info%trj_files(ti)%name == filename) then

        rep_in => cnv_info%trj_files(ti)
        call seek_trj(rep_in, trajectory, frm_no)
        return

      else
        cnv_info%trj_file_idx = cnv_info%trj_file_idx + 1
      end if

    end do

    ! open
    !
    do i = 1, MaxTrjFiles

      cnv_info%trj_file_idx = cnv_info%trj_file_idx + 1
      ti  = mod(cnv_info%trj_file_idx, MaxTrjFiles) + 1

      if (cnv_info%trj_files(ti)%name == '') &
        exit

    end do

    if (i > MaxTrjFiles) then
      cnv_info%trj_file_idx = cnv_info%trj_file_idx - 1
      ti = mod(cnv_info%trj_file_idx, MaxTrjFiles) + 1
    end if

    rep_in => cnv_info%trj_files(ti)

    if (i > MaxTrjFiles) &
      call close_trj(rep_in)

    write(MsgOut,'(A,A,A)') &
         '  Open trajectory file: "', trim(filename), '"'

    call open_trj(rep_in,         &
                  filename,       &
                  TrjFormatDCD,   &
                  TrjTypeCoorBox, &
                  IOFileInput)

    call seek_trj(rep_in, trajectory, frm_no)

    return

  end subroutine open_input_trj_

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    close_input_trj
  !> @brief        close all input trajectory file
  !! @authors      NT
  !! @param[inout] cnv_info   : information of convert
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine close_input_trj(cnv_info)

    ! formal arguments
    type(s_cnv_info),        intent(inout) :: cnv_info

    ! local variables
    integer                  :: i


    do i = 1, MaxTrjFiles
      if (cnv_info%trj_files(i)%name /= '') then
        call close_trj(cnv_info%trj_files(i))
        cnv_info%trj_files(i)%name = ''
      end if
    end do

    return

  end subroutine close_input_trj

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      get_replicate_name
  !> @brief        get replicate name
  !! @authors      NT
  !! @param[in]    filename   : filename
  !! @param[in]    replica_no : replica number
  !! @return       replicate name
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_replicate_name(filename, replica_no)

    ! return
    character(Maxfilename)   :: get_replicate_name

    ! formal arguments
    character(*),            intent(in)    :: filename
    integer,                 intent(in)    :: replica_no

    ! local variables
    integer                  :: bl, br
    character(MaxLine)       :: bl_str, br_str
    character(20)            :: fmt
    integer                  :: i, ndigit, comp


    bl = index(filename, '{', back=.true.)
    br = index(filename, '}', back=.true.)

    if (bl == 0 .or. br == 0) then
      get_replicate_name = filename
      return
    end if

    if (bl > br) &
      call error_msg('Get_Replicate_Name> Syntax error.')

    if (bl > 1) then
      bl_str = filename(1:bl-1)
    else
      bl_str = ''
    end if

    if (br < len_trim(filename)) then
      br_str = filename(br+1:)
    else
      br_str = ''
    end if
    
    do i = 1, 100
      comp = 10**i
      if (replica_no < comp) then
        ndigit = i
        exit
      end if
    end do
    
    write(fmt,'(a,i0,a,i0,a)') '(a,i', ndigit, '.', ndigit, ',a)'
    write(get_replicate_name,fmt) trim(bl_str), replica_no, trim(br_str)

    return

  end function get_replicate_name

end module rc_convert_mod
