!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   rc_option_mod
!> @brief   module for REMD trajectory conversion options
!! @authors Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module rc_option_mod

  use rc_option_str_mod
  use pbc_correct_mod
  use trajectory_str_mod
  use select_mod
  use select_atoms_mod
  use molecules_str_mod
  use fileio_control_mod
  use string_mod
  use messages_mod

  implicit none
  private

  ! structures
  type, public :: s_opt_info
    logical             :: check_only    = .true.
    integer             :: convert_type  = ConvertTypeParameter
    character(MaxLine)  :: convert_ids   = ''
    integer             :: dcd_md_period   = 0
    integer             :: trjout_format = TrjFormatDCD
    integer             :: trjout_type   = TrjTypeCoorBox
    integer             :: trjout_atom   = 1
    integer             :: pbc_correct   = PBCCModeNo
  end type s_opt_info

  public  :: show_ctrl_option
  public  :: read_ctrl_option
  public  :: setup_option
  private :: parse_convert_ids

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_option
  !> @brief        show control parameters in OPTION section
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_option

    write(MsgOut,'(A)') '[OPTION]'
    write(MsgOut,'(A)') 'check_only     = YES             # (YES/NO)'
    write(MsgOut,'(A)') 'convert_type   = PARAMETER       # (REPLICA/PARAMETER)'
    write(MsgOut,'(A)') 'convert_ids    =                 # (empty = all)(example: 1 2 5-10)'
    write(MsgOut,'(A)') 'dcd_md_period  = 0               # input dcdfile MD period'
    write(MsgOut,'(A)') 'trjout_format  = DCD             # (PDB/DCD)'
    write(MsgOut,'(A)') 'trjout_type    = COOR+BOX        # (COOR/COOR+BOX)'
    write(MsgOut,'(A)') 'trjout_atom    = 1'
    write(MsgOut,'(A)') 'pbc_correct    = NO              # (NO/MOLECULE)'
    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_option
  !> @brief        read control parameters in OPTION section
  !! @authors      NT
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   opt_info : OPTION section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_option(handle, opt_info)
  
    ! parameters
    character(*),            parameter     :: Section = 'Option'

    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_opt_info),        intent(inout) :: opt_info



    ! read parameters
    !

    call begin_ctrlfile_section(handle, Section)

    call read_ctrlfile_logical(handle, Section, 'check_only', &
                               opt_info%check_only)

    call read_ctrlfile_type   (handle, Section, 'convert_type', &
                               opt_info%convert_type, ConvertTypeTypes)

    call read_ctrlfile_string (handle, Section, 'convert_ids', &
                               opt_info%convert_ids)

    call read_ctrlfile_integer(handle, Section, 'dcd_md_period', &
                               opt_info%dcd_md_period)

    call read_ctrlfile_type   (handle, Section, 'trjout_format', &
                               opt_info%trjout_format, TrjFormatTypes)

    call read_ctrlfile_type   (handle, Section, 'trjout_type',  &
                               opt_info%trjout_type, TrjTypeTypes)

    call read_ctrlfile_integer(handle, Section, 'trjout_atom', &
                               opt_info%trjout_atom)

    call read_ctrlfile_type   (handle, Section, 'pbc_correct', &
                               opt_info%pbc_correct, PBCCModeTypes)

    call end_ctrlfile_section(handle)


    ! write parameters to MsgOut
    !

    write(MsgOut,'(A)') 'Read_Ctrl_Option> Parameters of Options'
    if (opt_info%check_only) then
      write(MsgOut,'(A20,A3)') '  check only      = ', 'yes'
    else
      write(MsgOut,'(A20,A2)') '  check only      = ', 'no'
    end if

    write(MsgOut,'(A20,A10)')  '  convert type    = ', &
         ConvertTypeTypes(opt_info%convert_type)
    write(MsgOut,'(A20,A10)')  '  convert ids     = ', &
         opt_info%convert_ids
    write(MsgOut,'(A20,I4)')   '  dcd md period   = ', &
         opt_info%dcd_md_period
    write(MsgOut,'(A20,A10)')  '  trjout format   = ', &
         TrjFormatTypes(opt_info%trjout_format)
    write(MsgOut,'(A20,A10)')  '  trjout type     = ', &
         TrjTypeTypes(opt_info%trjout_type)
    write(MsgOut,'(A20,A5,I0)')'  trjout atom     = ', &
         'group', opt_info%trjout_atom
    write(MsgOut,'(A20,A10)')  '  pbc correction  = ', &
         PBCCModeTypes(opt_info%pbc_correct)
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_option
  !> @brief        setup option information
  !! @authors      NT
  !! @param[in]    opt_info : OPTION section control parameters information
  !! @param[in]    sel_info : SELECTION section control parameters information
  !! @param[inout] molecule : molecule information
  !! @param[out]   option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_option(opt_info, &
                          sel_info, &
                          molecule, &
                          option)
  
    ! formal argments
    type(s_opt_info),        intent(in)    :: opt_info
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_molecule),        intent(inout) :: molecule
    type(s_option),          intent(inout) :: option


    ! check only
    option%check_only    = opt_info%check_only

    ! convert type
    option%convert_type  = opt_info%convert_type

    ! convert ids
    call parse_convert_ids(opt_info%convert_ids, option%convert_ids)

    ! dcd MD period
    option%dcd_md_period = opt_info%dcd_md_period

    ! trajectory format
    option%trjout_format = opt_info%trjout_format

    ! trajectory type
    option%trjout_type   = opt_info%trjout_type

    ! trajectory atom
    if (opt_info%trjout_atom > size(sel_info%groups)) &
      call error_msg('Setup_Option> trj-out atom selection is out of range.')

    option%trjout_atom_exp = sel_info%groups(opt_info%trjout_atom)

    call select_atom(molecule, &
                     option%trjout_atom_exp, &
                     option%trjout_atom)

    write(MsgOut,'(A,I8)') 'Setup_Option> trj-out atom count: ', &
         size(option%trjout_atom%idx)

    ! pbc correct
    option%pbcc_mode = opt_info%pbc_correct

    ! setup for PBC-correct mode "molecule"
    call setup_pbc_correct(option%pbcc_mode, molecule)

    write(MsgOut,'(A)') ' '

    return

  end subroutine setup_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    parse_convert_ids
  !> @brief        parse convert ids
  !! @authors      NT
  !! @param[in]    str     : ids information
  !! @param[out]   id_list : id list
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine parse_convert_ids(str, id_list)
  
    ! parameters
    integer,      parameter  :: MaxTok = 100

    ! formal arguments
    character(*),            intent(in)    :: str
    integer,    allocatable, intent(inout) :: id_list(:)
   
    ! local variables
    integer                  :: i, j, ntok, nids, rids, alloc_stat
    integer                  :: toki(2,MaxTok), ifr, ito
    character(20)            :: toks(  MaxTok)

    logical,    allocatable  :: ids(:)


    if (len_trim(str) == 0) return

    ! parse ids string
    toks(:) = ''
    read(str,*,end=10,err=900) toks

10  ntok = 0
    rids = 0

    do i = 1, MaxTok

      if (len_trim(toks(i)) == 0) &
        exit

      j = index(toks(i),'-')
      if (j /= 0) then

        ! range id
        toks(i)(j:j) = ' '
        read(toks(i),*,end=900,err=900) ifr, ito
        if (ifr > ito) ifr = ito

      else

        ! single id
        read(toks(i),*,end=900,err=900) ito
        ifr = ito

      end if

      toki(1,i) = ifr
      toki(2,i) = ito
      ntok = ntok + 1
      rids = MAX(rids, ito)

    end do

    ! check id 
    if (ntok == 0) &
      call error_msg('Parse_Convert_Ids> There is no IDs.')

    alloc_stat = 0
    allocate(ids(rids), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg('Parse_Convert_Ids> Id is too large.')
    ids(:) = .false.

    do i = 1, ntok
      do j = toki(1,i), toki(2,i)
        ids(j) = .true.
      end do
    end do

    nids = 0
    do i = 1, rids
      if (ids(i)) nids = nids + 1
    end do

    ! make id list
    allocate(id_list(nids), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    nids = 0
    do i = 1, rids
      if (ids(i)) then
        nids = nids + 1
        id_list(nids) = i
      end if
    end do

    ! deallocate
    deallocate(ids, stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_dealloc

    return

900 call error_msg('Parse_Convert_Ids> Read error. :'//trim(str))

  end subroutine parse_convert_ids

end module rc_option_mod
